#define _GNU_SOURCE

#include <errno.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <time.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>

#include <Conn.h>
#include <netstring.h>
#include <sesoma_conf.h>
#include <sesoma_pgsql.h>
#include <sesoma_mysql.h>


/* Global variables */
static unsigned short	debug = 1;

static FILE	*Logf = NULL;
static int	port = 8000;
static short	ipv4 = 1, ipv6 = 9;
static int	maxconn = 1000;

static void	(*sesoma_escape)(char *to, char *from, unsigned int from_len) = NULL;
static char	*(*sesoma_strerror)(void) = NULL;
static int	(*sesoma_connect)(char *host, char *port, char *user, char *pass, char *db, char *sock) = NULL;
static char	*(*sesoma_query)(char *query, int *ret) = NULL;


static void Log(unsigned short level, char *format, ...)
{
	va_list	ap;

	if (level > debug)
		return;

	fprintf(Logf, "[%ld] ", time(NULL));

	va_start(ap, format);
	vfprintf(Logf, format, ap);
	va_end(ap);
}

static void s_accept(Conn *C)
{
	Log(4, "%s (A connection was accepted from [%s] on slot %d. Set POLLIN and enq greeting...\n",
		__FUNCTION__, C->addr, C->slot);
}

static void s_close(Conn *C)
{
	Log(5, "%s (A connection will be closed [%s] on slot %d)\n",
		__FUNCTION__, C->addr, C->slot);
}

static unsigned int s_data(Conn *C)
{
	char tmp[8192], query[4096], escape1[4096];
	char answer[4096];
	int len, ret;
	char *buf, *buf2;
	char *map, *value, *x;

	len = C->ibuf_tail - C->ibuf_head;
	buf = C->ibuf + C->ibuf_head;

	if (debug >= 8) {
		Conn_dump(buf, len, tmp, sizeof(tmp));
		Log(8, "%s: recv head=%d tail=%d bytes on slot %d: [%s]\n",
			__FUNCTION__, C->ibuf_head, C->ibuf_tail,
			C->slot, tmp);
	}

	buf2 = netstring_decode(buf, len);
	if (buf2 == NULL) {
		Log(0, "Invalid/partial string received [%s]! Wait more.\n", buf);
		return 0;
	}

	Log(8, "	Received (after decode): [%s]\n", buf2);

	/* split */
	map = buf2;
	x = strchr(buf2, ' ');
	if (!x) {
		Log(0, "No space detected in request from sendmail buf=[%s] buf2=[%s]!\n",
			buf, buf2);
		Conn_close(C);
		return 0;
	}
	*x = '\0';
	value = x + 1;

	sesoma_escape(escape1, value, strlen(value));
	Log(0, "map=[%s], value=[%s]\n", map, value);

	/*
	if (strcmp(map, "access") == 0) {
		snprintf(answer, sizeof(answer), "OK OK");
	} else {
	*/
		snprintf(query, sizeof(query), conf_query, map, escape1);
		Log(0, "QUERY: %s\n", query);
		x = sesoma_query(query, &ret);
		switch (ret) {
			case 0: snprintf(answer, sizeof(answer), "OK %s", x); break;
			case 1: snprintf(answer, sizeof(answer), "NOTFOUND"); break;
			default: snprintf(answer, sizeof(answer), "TEMP Temporary error [%s]", sesoma_strerror()); break;
		}
	/*}*/

	Log(0, "Answer = [%s]\n", answer);
	x = netstring_encode(answer);
	Conn_enqueue(C, x, strlen(x));

	return len;
}

static void s_error(Conn *C)
{
	Log(1, "%s slot=%d C=%p\n",
		__FUNCTION__, C == NULL ? -1 : C->slot, (void *) C);
}

int main(void)
{
	Conn		*I4 = NULL, *I6 = NULL;
	int		ret;
	struct timeval	start;
	int		loops = 0;

	/* load config file */
	ret = sesoma_conf_load("/etc/sesoma/sesoma.conf");
	if (ret != 0) {
		fprintf(stderr, "Cannot load/parse /etc/sesoma/sesoma.conf [%s]!\n",
			sesoma_conf_strerror());
	}

	Logf = fopen(conf_log, "w");
	if (!Logf) {
		fprintf(stderr, "Cannot open log file [%s] [%s]\n",
			conf_log, strerror(errno));
		return 1;
	}
	chmod(conf_log, 0600);
	if (debug > 0)
		setlinebuf(Logf);


	if (getuid() == 0) {
		struct passwd *pw;
		struct group *g;
		uid_t uid;
		gid_t gid;

		/* switch group */
		g = getgrnam(conf_group);
		if (!g) {
			fprintf(stderr, "Cannot get info about group [%s] [%s]\n",
				conf_group, strerror(errno));
			return 1;
		}
		gid = g->gr_gid;
		ret = setresgid(gid, gid, gid);
		if (ret != 0) {
			fprintf(stderr, "Cannot switch to group [%s] [%s]\n",
				conf_user, strerror(errno));
			return 1;
		}

		/* switch user */
		pw = getpwnam(conf_user);
		if (!pw) {
			fprintf(stderr, "Cannot get info about user [%s] [%s]\n",
				conf_user, strerror(errno));
			return 1;
		}
		uid = pw->pw_uid;
		ret = setresuid(uid, uid, uid);
		if (ret != 0) {
			fprintf(stderr, "Cannot switch to user [%s] [%s]\n",
				conf_user, strerror(errno));
			return 1;
		}
	}


	daemon(0, 0);

	gettimeofday(&start, NULL);
	Log(0, "Starting at %ld...\n", start.tv_sec);
	Log(0, "\tPort=%d\n", port);
	Log(0, "\tLogFile=%s Debug=%d\n", conf_log, debug);
	Log(0, "\tDB: type=%s host=%s user=%s pass=%s db=%s port=%s sock=%s query=[%s]\n",
		conf_dbtype, conf_dbhost, conf_dbuser, conf_dbpass, conf_db, conf_dbport, conf_dbsock, conf_query);
	Log(0, "\tuser=%s group=%s\n", conf_user, conf_group);
	Log(0, "\tAfter switch: uid=%d euid=%d gid=%d\n", getuid(), geteuid(), getgid());
	Log(0, "\tipv4=%s ipv6=%s\n", (ipv4 == 1) ? "on" : "off", (ipv6 == 1) ? "on" : "off");

	/* set library output debug */
	Conn_debug(Logf, debug);

	/* db stuff */
	#ifdef SESOMA_PGSQL
	sesoma_escape = sesoma_pgsql_escape;
	sesoma_strerror = sesoma_pgsql_strerror;
	sesoma_connect = sesoma_pgsql_connect;
	sesoma_query = sesoma_pgsql_query;
	#endif
	#ifdef SESOMA_MYSQL
	sesoma_escape = sesoma_mysql_escape;
	sesoma_strerror = sesoma_mysql_strerror;
	sesoma_connect = sesoma_mysql_connect;
	sesoma_query = sesoma_mysql_query;
	#endif

	ret = sesoma_connect(conf_dbhost, conf_dbport, conf_dbuser, conf_dbpass, conf_db, conf_dbsock);
	if (ret != 0) {
		Log(0, "Cannot connect to database [%d] [%s]\n",
			ret, sesoma_strerror());
		Log(0, "I will try again later; till then I will temporary reject mails.\n");
	} else {
		Log(0, "Connection with db it's ok.\n");
	}

	Conn_default_ibuf = 256;
	Conn_default_obuf = 256;

	ret = Conn_init(maxconn);
	if (ret == -1) {
		Log(0, "%s", Conn_strerror());
		return 1;
	}

	if (ipv4 == 1) {
		Log(9, "Try to register IPv4 socket...\n");
		I4 = Conn_socket(PF_INET, SOCK_STREAM, port);
		if (!I4)
			Log(1, "Cannot bind on ipv4 socket [%s].\n", Conn_strerror());
	}

	if (ipv6 == 1) {
		Log(9, "Try to register IPv6 socket...\n");
		I6 = Conn_socket(PF_INET6, SOCK_STREAM, port);
		if (!I6)
			Log(1, "Cannot bind on ipv6 socket [%s].\n", Conn_strerror());
	}

	if ((I6 == NULL) && (I4 == NULL)) {
		Log(0, "Cannot bind!\n");
		return 1;
	}

	Conn_accept_cb = s_accept;
	Conn_data_cb = s_data;
	Conn_close_cb = s_close;
	Conn_error_cb = s_error;

	while (1) {
		ret = Conn_poll(10000);
		if (ret == -1) {
			Log(0, "Error calling Conn_poll [%s]!\n", Conn_strerror());
			return 1;
		}

		loops++;
		if (ret == 0)
			break;
	}

	Log(0, "Finish!\n\n");

	return 0;
}
